@import url('css/theme.css');

div.wy-nav-content {
    background-image: url('bluemarble.png');
    background-repeat: repeat-y;
    background-position: center;
}

div.diagram pre {
    font-family: monospace;
    font-size: 12px;
    padding: 5px;
    line-height: 95%;
    background: #fff;
    margin: 1px 0 24px 0;
}

div.figure p.caption {
    text-align: center;
}

div.figure p.caption:before {
    content: 'Figure: ';
    font-weight: bold;
}

p.homepage-header {
    width: auto;
    text-align: center;
    font-variant: small-caps;
    font-size: large;
    border-bottom: 1px solid #d3d3d3;
    padding-bottom: 2ex;
}
